/* $Copyright$
 *  Copyright (C) Fujitsu Siemens Computers GmbH 2001, 2002
 *  All rights reserved
 */
#ident "$Header$"
#ifndef WITHOUT_INCLUDES
#ifndef __KERNEL__
#  define __KERNEL__
#endif
#ifndef MODULE_SUB
#  define MODULE
#endif
#ifndef PRECOMPILE
#include <linux/modversions.h>
#endif
#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/string.h>
#include <linux/delay.h>
#include <linux/errno.h>
#include <linux/ioctl.h>
#include <linux/timer.h>
#include <linux/pci.h>
#include <asm/io.h>            	/* inb */
#include <asm/uaccess.h>		/* put_user */
#if LINUX_VERSION_CODE < 0x020400
#include <linux/kcomp.h>
#endif
#endif
#ident "$Header: $"
#ident "$Header$"
#ident "$Header: $"
typedef char CHAR;
typedef short SHORT;
typedef int BOOL;
typedef int INT;
typedef long LONG;
typedef float FLOAT;
typedef double DOUBLE;
typedef long long LONGLONG;
typedef unsigned char BYTE;
typedef unsigned char UCHAR;
typedef unsigned short WORD;
typedef unsigned short USHORT;
typedef unsigned int DWORD;
typedef unsigned int UINT;
typedef unsigned long ULONG;
typedef unsigned long long DWORDLONG;
typedef unsigned long long QWORD;
typedef DWORD time_t32;
typedef size_t SIZE_T;
typedef void VOID;
typedef CHAR *PCHAR;
typedef SHORT *PSHORT;
typedef FLOAT *PFLOAT;
typedef DOUBLE *PDOUBLE;
typedef UCHAR *PUCHAR;
typedef USHORT *PUSHORT;
typedef UINT *PUINT;
typedef ULONG *PULONG;
typedef CHAR *PSZ;
typedef CHAR *LPSZ;
typedef CHAR *LPSTR;
typedef CHAR *LPTSTR;
typedef CHAR *LPCSTR;
typedef CHAR *LPCTSTR;
typedef BOOL *PBOOL;
typedef BOOL *LPBOOL;
typedef BYTE *PBYTE;
typedef BYTE *LPBYTE;
typedef INT *PINT;
typedef INT *LPINT;
typedef WORD *PWORD;
typedef WORD *LPWORD;
typedef LONG *PLONG;
typedef LONG *LPLONG;
typedef DWORD *PDWORD;
typedef DWORD *LPDWORD;
typedef DWORDLONG *PDWORDLONG;
typedef DWORDLONG *LPDWORDLONG;
typedef VOID *PVOID;
typedef VOID *LPVOID;
typedef VOID *LPCVOID;
extern ULONG bmcBaseAddress;
extern UCHAR bmcRegSpacing;
typedef enum
{
        Unknown = 0x00,
        KCS,
        SMIC,
        BT
} InterfaceType, *PInterfaceType;
typedef struct
{
   BYTE data_reg;
   BYTE control_reg;
} SmicRegs;
typedef enum
{
    BYTE_BOUNDARY = 0x00,
    BIT32_BOUNDARY,
    BYTE16_BOUNDARY
}REGISTER_SPACING;
#pragma pack(1)
typedef struct tagSMBIOSEntryPoint
{
        char schAnchorString[4];
        BYTE byCheckSum;
        BYTE byLength;
        BYTE bySmBiosMajorVersion;
        BYTE bySmBiosMinorVersion;
        WORD wMaxSize;
        BYTE byepRevision;
        BYTE byFmtArea[5];
        BYTE byImAnchor[5];
        BYTE byImCheckSum;
        WORD wTableLength;
    DWORD dwTableAddress;
        WORD nSmBiosStructures;
        BYTE bybcdRevision;
} SMBIOSEntryPoint, *PSMBIOSEntryPoint;
typedef struct tadSMBIOSHeader
{
        BYTE Type;
        BYTE Length;
        WORD Handle;
} SMBIOSHeader, *PSMBIOSHeader;
typedef struct tagIPMIDeviceInformation
{
        SMBIOSHeader smbhdr;
    BYTE InterfaceType;
    BYTE SpecRevision;
    BYTE I2CSlaveAddress;
    BYTE NVDevAddress;
    struct {
                DWORD LowPart;
                DWORD HighPart;
        }BaseAddress;
    BYTE bBamIntrTrigger : 1;
    BYTE bBamIntrPolarity : 1;
    BYTE bBamIntrReserved : 1;
    BYTE bBamInterruptInfo : 1;
    BYTE bBamLSBitAddress : 1;
    BYTE bBamReserved : 1;
    BYTE bBamRegisterSpacing : 2;
    BYTE InterruptNumber;
} IPMIDeviceInformation, *PIPMIDeviceInformation;
typedef struct
{
    unsigned char cmdType;
    unsigned char rsSa;
    unsigned char busType;
    unsigned char netFn;
    unsigned char rsLun;
    unsigned char *data;
    int dataLength;
}
IMBPREQUESTDATA;
typedef struct
{
        IMBPREQUESTDATA request;
        unsigned char* preply;
        int replyLength;
        int timeout;
} IOCTL_IPMB_DATA;
#pragma pack()
typedef enum {
    ACCESS_OK,
    ACCESS_ERROR,
    ACCESS_OUT_OF_RANGE,
    ACCESS_END_OF_DATA,
    ACCESS_UNSUPPORTED,
    ACCESS_INVALID_TRANSACTION
}ACCESS_STATUS;
EXPORT_NO_SYMBOLS;
static const int SUCCESS = 0;
static const char *IPMI_DEV = "ipmi";
static unsigned char status;
static int ipmi_deviceOpen = -1;
static int ipmi_major = 0;
static int sequence=0;
static BOOL OldIpmi=0;
static InterfaceType Interface = Unknown;
static ULONG bmcBaseAddress;
static UCHAR bmcRegSpacing = 1;
static unsigned int bmcTimeoutOccured = 0;
static struct timer_list bmcTimer;
static short ipmiDebug = 0;
MODULE_PARM(ipmiDebug, "h");
static void ipmi_setBMCtimeout(unsigned long data) {
        bmcTimeoutOccured = 1;
        if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "-->ipmi_setBMCtimeout: BMC Timer expired\n");
        return;
}
BOOL
ScanForIPMIDeviceInfo (
 ULONG paSmBiosTableAddress,
 ULONG ulSmBiosTableLength
        )
{
    int i;
    ULONG Length = ulSmBiosTableLength;
    PUCHAR pVirtualAddress = NULL, pucTemp = NULL;
        BOOL ErrataCompliant = 0;
        BOOL Success = 0;
        PIPMIDeviceInformation pInfo = NULL;
        if ((!paSmBiosTableAddress) || (ulSmBiosTableLength <= 0)) {
                if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "Invalid Physical Address and Range of Bios Table\n");
                return 0;
        }
        pVirtualAddress = phys_to_virt(paSmBiosTableAddress);
        if (!pVirtualAddress) {
                if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "No Virtual Pointer to SMBios Table\n");
                return 0;
        }
        if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "INFO: SMBios Table 0x%0lX Length: 0x%0lX\n", paSmBiosTableAddress, ulSmBiosTableLength);
        while (Length > 0) {
                if (((PSMBIOSHeader)pVirtualAddress)->Type == 38) {
                        pInfo = (PIPMIDeviceInformation)pVirtualAddress;
                        if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "Located IPMI Device Info Record @0x%p Len: 0x%02X\n", (void *)pVirtualAddress, ((PSMBIOSHeader)pVirtualAddress)->Length);
                        switch(pInfo->InterfaceType) {
                                case KCS:
                                        if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "KCS Interface ...");
                                        Interface = KCS;
                                        break;
                                case SMIC:
                                        if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "SMIC Interface ...");
                                        Interface = SMIC;
                                        break;
                                case Unknown:
                                case BT:
                                default:
                                        if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "BT/Unknown Interface NOT supported.\n");
                                        Interface = Unknown;
                                        break;
                        }
                        bmcRegSpacing = 1;
                        if ( ((PSMBIOSHeader)pVirtualAddress)->Length > 0x10) {
                                ErrataCompliant = 1;
                        }
                        if (ErrataCompliant) {
                                switch(pInfo->bBamRegisterSpacing)
                                {
                                        case BYTE_BOUNDARY:
                                                bmcRegSpacing = 1;
                                                if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "Reg. Spacing is BYTE (1)\n");
                                                break;
                                        case BIT32_BOUNDARY:
                                                if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "Reg. Spacing is DWORD (4)\n");
                                                bmcRegSpacing = 4;
                                                break;
                                        case BYTE16_BOUNDARY:
                                                if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "Reg. Spacing is QWORD (16)\n");
                                                bmcRegSpacing = 16;
                                                break;
                                        default:
                                                if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "Inv. Register Spacing ... using default");
                                                bmcRegSpacing = 1;
                                                break;
                                }
                        }
                        if (pInfo->BaseAddress.LowPart & 0x01) {
                                bmcBaseAddress = pInfo->BaseAddress.LowPart;
                                if (ErrataCompliant) {
                                        bmcBaseAddress &= ~0x01;
                                        bmcBaseAddress |= pInfo->bBamLSBitAddress;
                                }
                                if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "Final Base Address is @ 0x%lX\n", bmcBaseAddress);
                                switch (Interface) {
                                        case KCS:
                                                if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "INFO: KCS DATA: 0x%lX KCS CMD_STATUS: 0x%lX \n", (bmcBaseAddress), ((bmcBaseAddress) + 1));
                                                break;
                                        case SMIC:
                                                if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "INFO: SMIC DATA: 0x%lX  CNTL: 0x%lX FLAG: 0x%lX\n", (bmcBaseAddress), ((bmcBaseAddress) + 1), ((bmcBaseAddress) + 2));
                                                break;
                                        default:
                                                if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "INFO: Unsupported Interface in Record 38\n");
                                                break;
                                }
                        }
                        Success = 1;
                        break;
                }
                Length -= ((PSMBIOSHeader)pVirtualAddress)->Length;
                pVirtualAddress += ((PSMBIOSHeader)pVirtualAddress)->Length;
                for(i=0,pucTemp = pVirtualAddress;(ULONG)i < Length;i++) {
                        if (!(pucTemp[i] || pucTemp[i+1])) {
                                break;
                        }
                }
                Length -= (i+2);
                pVirtualAddress += (i+2);
        }
    return Success;
}
BOOL
GetIPMIDeviceInformation (void)
{
        BOOL Success = 0;
        ULONG SmBiosAddress = 0xF0000;
    ULONG ulLength = 0xFFFE;
        ULONG SmTableAddress = 0;
    ULONG ulTableLength = 0;
        PSMBIOSEntryPoint pSmEp = NULL;
        UCHAR ucChecksum;
    PUCHAR pVirtualAddress = NULL;
    ULONG Offset;
        pVirtualAddress = phys_to_virt(SmBiosAddress);
        if (!pVirtualAddress) {
                if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "ERROR: No Virtual pointer to SMBIOS Memory\n");
                return 0;
        }
        for(Offset=0,pSmEp=NULL; Offset<ulLength - sizeof("_SM_") + 1; Offset++) {
        if (strncmp(pVirtualAddress + Offset, "_SM_", sizeof("_SM_")-1) == 0) {
                        pSmEp = (PSMBIOSEntryPoint)(pVirtualAddress + Offset);
                        if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "SMBIOS Entry Point @ 0x%p, offset: 0x%0lX\n", (void *)pSmEp, Offset);
                        break;
                }
        }
        if (!pSmEp) {
                if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "WARNING: Unable to Locate SMBIOS signature\n");
                return 0;
        }
        for(Offset=0,ucChecksum=0x00;Offset<pSmEp->byLength;Offset++) {
                ucChecksum += (UCHAR) *((PUCHAR)pSmEp + Offset);
        }
        if (ucChecksum) {
                if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "Invalid BIOS Entrypoint Checksum\n");
                return 0;
        }
        if ((pSmEp->bySmBiosMajorVersion < 2) || (pSmEp->bySmBiosMinorVersion < 3)) {
                if (ipmiDebug) printk(KERN_DEBUG "ipmi :" " SMBIOS Incompatible Version (%x.%x) detected\n", pSmEp->bySmBiosMajorVersion, pSmEp->bySmBiosMinorVersion);
        } else {
                SmTableAddress = (ULONG)pSmEp->dwTableAddress;
                ulTableLength = (ULONG)pSmEp->wTableLength;
                Success = ScanForIPMIDeviceInfo(SmTableAddress, ulTableLength);
                if (!Success) {
                        if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "ERROR: Record Type 38 Could NOT be located\n");
                }
        }
    return Success;
}
BOOL IsISA(void)
{
   if((inb(((bmcBaseAddress) + 1)) == 0xFF) &&
      (inb((bmcBaseAddress)) == 0xFF))
   {
      return(0);
   }
   else
   {
      return(1);
   }
}
BOOL
test_for_obf()
{
   status = ((BYTE)(inb(((bmcBaseAddress) + 1))));
   return ((status & 0x01) == 0x01);
}
int wait_for_obf(int timeout)
{
   int i = 0;
   do {
      if (test_for_obf()) {
                  udelay (5);
                  return (0);
          }
      i++;
      set_current_state(TASK_UNINTERRUPTIBLE);
      schedule_timeout((1*HZ)/1000);
   } while (i < timeout);
   if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "wait_while_obf ERROR (timeout was %d Status: 0x%02X\n", timeout, ((BYTE)(inb(((bmcBaseAddress) + 1)))));
   return( -1);
}
BOOL
test_for_ibf()
{
   status = ((BYTE)(inb(((bmcBaseAddress) + 1))));
   return ((status & 0x02) == 0x02);
}
int wait_while_ibf(int timeout)
{
   int i = 0;
   do {
      if (0 == test_for_ibf()) {
                  udelay (5);
                  return (0);
          }
      i++;
      set_current_state(TASK_UNINTERRUPTIBLE);
      schedule_timeout((1*HZ)/1000);
   } while (i < timeout);
   if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "wait_while_ibf ERROR (timeout was %d Status: 0x%02X\n", timeout, ((BYTE)(inb(((bmcBaseAddress) + 1)))));
   return( -1);
}
int ISASendMessage(unsigned char *msgBuf, int length, int timeout)
{
        int i;
        if (wait_while_ibf(1000))
        {
                if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "Timeout waiting while ibf full\n");
                return(-1);
        }
        (outb(0x61,((bmcBaseAddress) + 1)));
        if (wait_while_ibf(50))
        {
                if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "Timeout during WRITE_START ibf full\n");
                return(-1);
        }
        for (i = 0; i < length - 1; i++){
                udelay(20);
                status = ((BYTE)(inb(((bmcBaseAddress) + 1))));
                if (status & 0x01){
                        ((BYTE)(inb((bmcBaseAddress))));
                }
                if ((status & 0xC0) != 0x80){
                        if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "ERROR invalid KCS state 0x%02X on ReqByte %d\n", (status & 0xC0), i);
                        return(-1);
                }
                (outb(msgBuf[i],(bmcBaseAddress)));
                if (wait_while_ibf(500)){
                        if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "Timeout during DATA ibf full\n");
                        return(-1);
                }
        }
        (outb(0X62,((bmcBaseAddress) + 1)));
        if (wait_while_ibf(50)){
                if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "Timeout during WRITE_END ibf full\n");
                return(-1);
        }
        udelay(10);
        status = ((BYTE)(inb(((bmcBaseAddress) + 1))));
        if ((status & 0xC0) != 0x80){
                if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "ERROR invalid KCS state 0x%02X on WRITE_END\n", (status & 0xC0));
                return(-1);
        }
        if (status & 0x01) {
                ((BYTE)(inb((bmcBaseAddress))));
        }
        (outb(msgBuf[i],(bmcBaseAddress)));
        if (wait_while_ibf(50)){
                if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "Timeout during last DATA ibf full\n");
                return(-1);
        }
        udelay(20);
        status = ((BYTE)(inb(((bmcBaseAddress) + 1))));
        if ((status & 0xC0) != 0x40){
                if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "ERROR invalid KCS state 0x%02X expect READ_STATE\n", (status & 0xC0));
                return(-1);
        }
        return(0);
}
int ISAReadBMCData(unsigned char *msgBuf, int timeout)
{
        int i;
        BYTE state;
        BYTE dummy;
        for (i = 0; i < 35; ) {
                do {
                        udelay(20);
                        status = ((BYTE)(inb(((bmcBaseAddress) + 1))));
                        state = (BYTE)(status & 0xC0);
                        switch (state) {
                                case 0xC0:
                                        if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "ISAReadBMCData invalid ISA_ERROR_STATE\n");
                                        return(-1);
                                        break;
                                case 0x80:
                                        if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "ISAReadBMCData invalid ISA_WRITE_STATE\n");
                                        return(-1);
                                        break;
                                case 0x40:
                                        if ((status & 0x01) != 0x01) {
                            set_current_state(TASK_UNINTERRUPTIBLE);
                            schedule_timeout((1*HZ)/1000);
                        timeout--;
                                        }
                                        if (timeout == 0) {
                                                if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "WARNING: Timeout waiting for obf when READ_STATE signalled\n");
                                                if (i >= 3)
                                                        return(i);
                                                else
                                                        return(-1);
                                        }
                                        break;
                                case 0x00:
                                        if (OldIpmi) {
                                                return(i);
                                        } else {
                                                if (wait_for_obf(50)) {
                                                        if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "WARNING: Timeout waiting for FINAL obf (IPMI 1.0 or later)\n");
                                                }
                                                dummy = ((BYTE)(inb((bmcBaseAddress))));
                                                return(i);
                                        }
                                        break;
                                default:
                                        return(-1);
                        }
                } while ((status & 0x01) != 0x01);
                udelay(10);
                msgBuf[i] = ((BYTE)(inb((bmcBaseAddress))));
                i++;
                udelay(10);
                if (wait_while_ibf(50)) {
                        if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "ISAReadBMCData wait_while_ibf\n");
                        return(-1);
                }
                (outb(0X68,(bmcBaseAddress)));
        }
        if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "ISAReadBMCData ISA_IDLE_STATE ERROR\n");
        return(-1);
}
BOOL IsSMIC(void)
{
   if((inb(((bmcBaseAddress) + 1)) != 0xFF))
   {
      return(1);
   }
   else
   {
      return(0);
   }
}
static int wait_for_condition (int cond, int delay_loops, int busy_loops)
{
   int dl, bl;
   if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "wait_for_condition %d, delay_loops=%d, busy_loops=%d\n", cond,delay_loops,busy_loops);
   for (dl=0; dl < delay_loops; dl ++) {
       switch (cond) {
         case 1 :
              for (bl=0; bl<busy_loops; bl++) {
                  if (((inb(((bmcBaseAddress) + 2)) & 0x01) == 0x00)) {
                      if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "wait_4_not_busy dloops=%d, bloops=%d\n", dl,bl);
                      return (0);
                  }
              }
              break;
         case 2 :
              for (bl=0; bl<busy_loops; bl++) {
                  if (((inb(((bmcBaseAddress) + 2)) & (0x01 | 0x80)) == 0x80)) {
                      if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "wait_4_data_ready dloops=%d, bloops=%d\n", dl,bl);
                      return (0);
                  }
              }
              break;
         case 3 :
              for (bl=0; bl<busy_loops; bl++) {
                  if (((inb(((bmcBaseAddress) + 2)) & (0x01 | 0x40)) == 0x40)) {
                      if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "wait_4_send_ready dloops=%d, bloops=%d\n", dl,bl);
                      return (0);
                  }
              }
              break;
         default:
              if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "--> ERROR wait_for_condition: EINVAL\n");
              return (-EINVAL);
       }
       set_current_state(TASK_UNINTERRUPTIBLE);
       schedule_timeout(0);
       if (bmcTimeoutOccured == 1) {
              if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "wait_for_condition: BMC Timer expired\n");
          if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "wait_for_condition: %d dloops\n", dl);
          return (-ETIME);
       }
    }
    if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "wait_for_condition: %d dloops\n", dl);
    if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "--> wait_for_condition: MAX-TIMEOUT occured\n");
    return (-ETIME);
}
static unsigned int wait_cond_transmit_ready = 3;
static void test_TX_READY_flag (void)
{
   unsigned int timeout;
   timeout = 300 * 5;
   init_timer(&bmcTimer);
   bmcTimer.function = &ipmi_setBMCtimeout;
   bmcTimer.data = (unsigned long) 0;
   bmcTimer.expires = jiffies + ((unsigned long )timeout*HZ)/1000;
   bmcTimeoutOccured = 0;
   add_timer(&bmcTimer);
   if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "test_TX_READY_flag: BMC timer started for %d ms\n", timeout);
   if (wait_for_condition(3,10000,100) != 0) {
       if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "WAIT_COND_TRANSMIT_READY is NOT signaled\n");
       wait_cond_transmit_ready = 1;
   } else {
       if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "WAIT_COND_TRANSMIT_READY has been signaled\n");
       wait_cond_transmit_ready = 3;
   }
   del_timer_sync(&bmcTimer);
   if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "test_TX_READY_flag: BMC timer stoped\n");
   return;
}
SmicRegs SmicOut(BYTE smic_data, BYTE smic_control)
{
   SmicRegs getc_regs;
   UCHAR flagReg;
   outb(smic_data,(bmcBaseAddress));
   outb(smic_control,((bmcBaseAddress) + 1));
   flagReg = inb(((bmcBaseAddress) + 2));
   outb((UCHAR)(flagReg | 0x01),((bmcBaseAddress) + 2));
   if (wait_for_condition(1,10000,1000) != 0) {
           if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "WAIT_COND_NOT_BUSY is NOT Signaled\n");
           getc_regs.control_reg=0;
       return(getc_regs);
   }
   getc_regs.data_reg= (BYTE)(inb((bmcBaseAddress)));
   getc_regs.control_reg= (BYTE)(inb(((bmcBaseAddress) + 1)));
   return(getc_regs);
}
int do_SMICSendMessage(unsigned char *msgBuf, int length)
{
   SmicRegs netcom;
   int i;
   int ret;
   if ((ret=wait_for_condition(wait_cond_transmit_ready,10000,100)) != 0) {
      return (ret);
   }
   netcom = SmicOut(*msgBuf++,0x41);
   if (netcom.control_reg != 0xC1)
   {
      if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "ERROR - did not receive SC_SMS_WR_START from SMIC\n");
      return (-1);
   }
   for (i = 0; i < length - 2; i++)
   {
      if ((ret=wait_for_condition(wait_cond_transmit_ready,10000,100)) != 0) {
         return (ret);
      }
      netcom = SmicOut(*msgBuf++,0x42);
      if (netcom.control_reg != 0xC2)
      {
         if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "ERROR - did not receive SC_SMS_WR_NEXT from SMIC\n");
         return(-1);
      }
   }
   if ((ret=wait_for_condition(wait_cond_transmit_ready,10000,100)) != 0) {
      return (ret);
   }
   netcom = SmicOut(*msgBuf,0x43);
   if (netcom.control_reg != 0xC3)
   {
      if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "ERROR - did not receive SC_SMS_WR_END from SMIC\n");
      return(-1);
   }
   return(0);
}
int do_SMICReadBMCData(unsigned char *msgBuf)
{
   SmicRegs netcom;
   int count = 0;
   int ret;
   if ((ret=wait_for_condition(2,10000,10)) != 0) {
      return (ret);
   }
   netcom = SmicOut(0,0x44);
   if (netcom.control_reg != 0xC4)
   {
      if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "Failed to get Read Start from SMIC.\n");
      return(-1);
   }
   else
   {
      *msgBuf++ = netcom.data_reg;
   }
   for (count = 1; count < 67; count++)
   {
      if ((ret=wait_for_condition(2,10000,10)) != 0) {
         return (ret);
      }
      netcom = SmicOut(0,0x45);
      if (netcom.control_reg != 0xC5)
      {
         break;
      }
      else
      {
         *msgBuf++ = netcom.data_reg;
      }
   }
   if (netcom.control_reg != 0xC6)
   {
          if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "SMIC_SC_SMS_RD_END not received\n");
      return(-1);
   }
   else
   {
      *msgBuf++ = netcom.data_reg;
   }
   if ((ret=wait_for_condition(2,10000,10)) != 0) {
      return (ret);
   }
   netcom = SmicOut(0,0x46);
   if ((netcom.control_reg != 0xC0) ||
     (netcom.data_reg != 0) || (count == 0))
   {
          if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "ERROR !!!!\n");
      return(-1);
   }
   return (count+1);
}
int SMICReadBMCData(unsigned char *msgBuf, int timeout)
{
    int ret = 0;
   init_timer(&bmcTimer);
   bmcTimer.function = &ipmi_setBMCtimeout;
   bmcTimer.data = (unsigned long) 0;
   bmcTimer.expires = jiffies + ((unsigned long )timeout*HZ)/1000;
   bmcTimeoutOccured = 0;
   add_timer(&bmcTimer);
   if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "SMICReadBMCData: BMC timer started for %d ms\n", timeout);
   ret = do_SMICReadBMCData(msgBuf);
   del_timer_sync(&bmcTimer);
   if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "SMICReadBMCData: BMC timer stoped\n");
   return (ret);
}
int SMICSendMessage(unsigned char *msgBuf, int length, int timeout)
{
    int ret = 0;
   init_timer(&bmcTimer);
   bmcTimer.function = &ipmi_setBMCtimeout;
   bmcTimer.data = (unsigned long) 0;
   bmcTimer.expires = jiffies + ((unsigned long )timeout*HZ)/1000;
   bmcTimeoutOccured = 0;
   add_timer(&bmcTimer);
   if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "SMICSendMessage: BMC timer started for %d ms\n", timeout);
   ret = do_SMICSendMessage(msgBuf,length);
   del_timer_sync(&bmcTimer);
   if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "SMICSendMessage: BMC timer stoped\n");
   return (ret);
}
void DetermineInterface(void)
{
        if (Interface == Unknown) {
                if (!GetIPMIDeviceInformation()) {
                        bmcBaseAddress = ((ULONG)0xCA2);
                        if(IsISA()) {
                                if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "BMC Interface = Legacy KCS\n");
                                Interface = KCS;
                        } else {
                                bmcBaseAddress = ((ULONG)0x0CA9);
                                if (IsSMIC()) {
                                        if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "BMC Interface = Legacy SMIC\n");
                                        Interface = SMIC;
                                        test_TX_READY_flag();
                                } else {
                                        if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "Could not detect BMC Interface - NON IPMI system\n");
                                        Interface = Unknown;
                                }
                        }
                }
        }
}
int ReadBMCData(unsigned char *msgBuf, int timeout)
{
   switch(Interface)
   {
      case KCS:
         return(ISAReadBMCData(msgBuf, timeout));
      case SMIC:
         return(SMICReadBMCData(msgBuf, timeout));
      case BT:
      case Unknown:
         if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "Error - unable to determine interface type\n");
   }
   return(0);
}
int SendMessage(unsigned char *msgBuf, int length, int timeout)
{
   switch(Interface)
   {
      case KCS:
         return(ISASendMessage(msgBuf,length, timeout));
      case SMIC:
         return(SMICSendMessage(msgBuf,length, timeout));
      case BT:
      case Unknown:
         if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "Error - unable to determine interface type\n");
   }
   return(0);
}
int IsIPMIControllerAvailable(void)
{
   unsigned char CmdBuffer[256];
   unsigned char ResponseBuffer[256];
   int BytesRead;
   int i;
   CmdBuffer[0] = 0x18;
   CmdBuffer[1] = 0x01;
   DetermineInterface();
   if(SendMessage(CmdBuffer, 0x02, 300) != 0)
   {
      if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "Error sending GetDeviceID command to BMC.\n");
          return 0;
   }
   else
   {
          BytesRead = ReadBMCData(ResponseBuffer, 300);
      if (BytesRead < 0) {
          if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "ReadBMCData returned an error %d\n", BytesRead);
          return 0;
      }
          if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "OK sending GetDeviceID command to BMC. received %d bytes\n", BytesRead);
          for (i=0;i<BytesRead;i++)
          {
             if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "Byte %d 0x%02X\n", i,ResponseBuffer[i]);
          }
          if ((BytesRead >= 8) && (ResponseBuffer[2] == 0))
          {
                OldIpmi = (BytesRead < 14) ? 1 : 0;
                if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "Detected %s IPMI Interface\n", OldIpmi ? "OLD (0.9)" : "NEW (1.0 or later)");
                return 1;
          }
          else
                return 0;
   }
}
static int calcChecksum(unsigned char *pData,int length)
{
        int i;
        int checksum;
        checksum=0;
        for(i=0;i<length;i++,pData++)
        {
                checksum += *pData;
        }
        checksum=(~checksum) + 1;
        return (checksum);
}
static int ipmi_ioctl(
                                struct inode *inode,
                                struct file *file,
                                unsigned int cmd,
                                unsigned long arg )
{
        int i,retries;
        if (_IOC_TYPE(cmd) != 'i') {
                if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "  type != magic\n");
                return -EINVAL;
        }
        if (_IOC_NR(cmd) > 1) {
                if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "  nr > maxnr\n");
                return -EINVAL;
        }
        switch (cmd) {
        case _IOR ('i', 0, IOCTL_IPMB_DATA):
                {
                        int bytesread;
                        IOCTL_IPMB_DATA data;
                        unsigned char msgbuffer[256];
                        unsigned char replybuffer[256];
                        copy_from_user(&data,(IOCTL_IPMB_DATA *)arg,sizeof(IOCTL_IPMB_DATA));
                        if (data.request.rsSa != 0x20) {
                                if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "IPMB Request I2C: 0x%02X Cmd 0x%02X NetFn: 0x%02X LUN %d Len %d\n", data.request.rsSa, data.request.cmdType, data.request.netFn, data.request.rsLun, data.request.dataLength);
                                msgbuffer[0]=0x18;
                                if (OldIpmi) {
                                        msgbuffer[1]=0x50;
                                } else {
                                        msgbuffer[1]=0x34;
                                }
                                msgbuffer[2]=data.request.busType;
                                msgbuffer[3]=data.request.rsSa;
                                msgbuffer[4]=(((data.request.netFn << 2) & 0xFC) | (data.request.rsLun & 3));
                                msgbuffer[5]=calcChecksum(&msgbuffer[3],2);
                                msgbuffer[6]=0x20;
                                sequence++;
                                msgbuffer[7]=(((sequence << 2) & 0xFC) | 0x02);
                                msgbuffer[8]=data.request.cmdType;
                                if (data.request.dataLength > 0)
                                {
                                        if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "Cmd 0x%x Request datalength %d\n", data.request.cmdType,data.request.dataLength);
                                        if ((data.request.dataLength < (256-2)) && (data.request.data))
                                        {
                                                copy_from_user(&msgbuffer[9],data.request.data,data.request.dataLength);
                                        }
                                }
                                msgbuffer[9+data.request.dataLength]=calcChecksum(&msgbuffer[6],3+data.request.dataLength);
                                for (i=0;i<10;i++)
                                        if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "data %d: 0x%x\n", i,msgbuffer[i]);
                                if(SendMessage(msgbuffer, data.request.dataLength + 10, data.timeout) != 0)
                                {
                                        if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "Error sending command Write Master SendMessage to BMC.\n");
                                        return -EINVAL;
                                }
                                bytesread = ReadBMCData(replybuffer, data.timeout);
                                if (bytesread < 0) {
                                    if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "ReadBMCData returned an error %d\n", bytesread);
                                    return (bytesread);
                                }
                                if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "ReadBMCData read %d bytes\n", bytesread);
                                if (replybuffer[2] == 0)
                                {
                                        for(retries=0;retries<20;retries++)
                                        {
                                                msgbuffer[0]=0x18;
                                                if (OldIpmi) {
                                                        msgbuffer[1]=0x37;
                                                } else {
                                                        msgbuffer[1]=0x33;
                                                }
                                                if(SendMessage(msgbuffer,2, data.timeout) != 0)
                                                {
                                                        if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "Error sending GetMessage command to BMC.\n");
                                                        continue;
                                                }
                                                bytesread = ReadBMCData(replybuffer, data.timeout);
                                                if (bytesread < 0) {
                                                    if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "ReadBMCData returned an error %d\n", bytesread);
                                                    continue;
                                                }
                                                for (i=0;i<bytesread;i++)
                                                        if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "reply data %d: 0x%x\n", i,replybuffer[i]);
                                                if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "ReadBMCData read %d bytes CompletionCode 0x%x\n", bytesread,replybuffer[2]);
                                                if (replybuffer[2] == 0)
                                                {
                                                        if ((replybuffer[7] == data.request.cmdType) && ((sequence & 0x3F) == ((replybuffer[6] >> 2) & 0x3F)))
                                                        {
                                                                break;
                                                        }
                                                        else
                                                        {
                                                                if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "ERROR Got SMS Message which is not our response\n");
                                                        }
                                                }
                                        }
                                        if (retries >= 20)
                                        {
                                                if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "Could not get SMS Message after %d retries\n", 20);
                                                return -EINVAL;
                                        }
                                        if ((bytesread > 7) && (data.preply) && (bytesread < data.replyLength))
                                        {
                                                data.replyLength=bytesread-7;
                                                copy_to_user(data.preply,&replybuffer[6],data.replyLength);
                                                copy_to_user((IOCTL_IPMB_DATA *)arg,&data,sizeof(IOCTL_IPMB_DATA));
                                                return 0;
                                        }
                                        else
                                        {
                                                if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "ReadBMCData SMS Message too small %d bytes read\n", bytesread);
                                                return -EINVAL;
                                        }
                                }
                                else
                                {
                                        if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "ReadBMCData CompletionCode SendMessage 0x%x 0x%x 0x%x \n", replybuffer[0],replybuffer[1],replybuffer[2]);
                                        return -EINVAL;
                                }
                        }
                        else
                        {
                                if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "BMC Req Cmd 0x%02X NetFn 0x%02X LUN %d Len %d\n", data.request.cmdType, data.request.netFn, data.request.rsLun, data.request.dataLength);
                                msgbuffer[0]=(((data.request.netFn << 2) & 0xFC) | (data.request.rsLun & 3));
                                msgbuffer[1]=data.request.cmdType;
                                if (data.request.dataLength > 0) {
                                        if ((data.request.dataLength < (256-2)) && (data.request.data)) {
                                                copy_from_user(&msgbuffer[2],data.request.data,data.request.dataLength);
                                        } else {
                                                if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "ERROR Invalid data pointer\n");
                                                return -EINVAL;
                                        }
                                }
                                if(SendMessage(msgbuffer, data.request.dataLength + 2, data.timeout) != 0) {
                                        if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "Error sending command to BMC.\n");
                                        return -EINVAL;
                                }
                                bytesread = ReadBMCData(replybuffer, data.timeout);
                                if (bytesread < 0) {
                                    if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "ReadBMCData returned an error %d\n", bytesread);
                                    return (bytesread);
                                }
                                if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "ReadBMCData returned %d bytes\n", bytesread);
                                if ((bytesread > 0) && (data.preply) && (bytesread < data.replyLength)) {
                                        data.replyLength=bytesread;
                                        copy_to_user(data.preply,replybuffer,data.replyLength);
                                        copy_to_user((IOCTL_IPMB_DATA *)arg,&data,sizeof(IOCTL_IPMB_DATA));
                                        return 0;
                                } else {
                                        if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "Error invalid reply pointer\n");
                                        return -EINVAL;
                                }
                        }
                }
        default:
                                if (ipmiDebug) printk(KERN_DEBUG "ipmi :" " ioctl-cmd=%x not valid\n", cmd);
                                return -EINVAL;
        }
        return SUCCESS;
}
static int ipmi_open(struct inode *inode, struct file *file)
{
    if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "-->ipmi_open: (%p,%p)\n", (void *)inode, (void *)file);
    if(ipmi_deviceOpen < 0) {
        if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "ipmi_open: device not initialized\n");
        return -ENODEV;
    }
    ipmi_deviceOpen++;
    MOD_INC_USE_COUNT;
    return SUCCESS;
}
static int ipmi_release(struct inode *inode, struct file *file)
{
    del_timer_sync(&bmcTimer);
    ipmi_deviceOpen--;
    if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "-->ipmi_release: (%p,%p)\n", (void *)inode, (void *)file);
    MOD_DEC_USE_COUNT;
    return SUCCESS;
}
static struct file_operations ipmi_fops =
{
    .ioctl = ipmi_ioctl,
    .open = ipmi_open,
    .release = ipmi_release
};
int init_module(void)
{
    if (! IsIPMIControllerAvailable()) {
        if (ipmiDebug) printk(KERN_DEBUG "ipmi :" "init_module: IPMI not present\n");
        return -ENODEV;
    }
    if((ipmi_major=register_chrdev(ipmi_major, IPMI_DEV, &ipmi_fops)) < 0) {
        printk(KERN_ERR "ipmi: unable to get major %d\n", ipmi_major);
        return -EIO;
    }
    ipmi_deviceOpen = 0;
    return SUCCESS;
}
void cleanup_module(void)
{
    int err;
    del_timer_sync(&bmcTimer);
    err = unregister_chrdev(ipmi_major, IPMI_DEV);
    if (err < 0) {
        printk(KERN_ERR "ipmi: unregister_chrdev(%i) failed!\n", ipmi_major);
    }
    ipmi_deviceOpen = -1;
}
